"""
A few generally useful string- and file-related utilities.
"""

__author__ = 'jlevy'

from string import Template
import os
import random
import shutil
import shlex
import pipes
import hashlib
from contextlib import contextmanager
from datetime import datetime

# The subprocess module has known threading issues, so prefer subprocess32.
try:
  import subprocess32 as subprocess
except ImportError:
  import subprocess

DEV_NULL = open(os.devnull, 'wb')

BACKUP_SUFFIX = ".bak"

_RANDOM = random.SystemRandom()
_RANDOM.seed()


def new_uid(bits=64):
  """
  A random alphanumeric value with at least the specified bits of randomness. We use base 36,
  i.e. not case sensitive. Note this makes it suitable for filenames even on case-insensitive disks.
  """
  return "".join(_RANDOM.sample("0123456789abcdefghijklmnopqrstuvwxyz",
                                int(bits / 5.16) + 1))  # log(26 + 10)/log(2) = 5.16


def iso_timestamp():
  """
  ISO timestamp.
  """
  return datetime.now().isoformat() + 'Z'


def dict_merge(*dict_list):
  '''
  Given zero or more dicts, shallow copy and merge them into a new dict, with
  precedence to dictionary values later in the dict list.
  '''
  result = {}
  for d in dict_list:
    result.update(d)
  return result


#
# ---- Templating ----

def expand_variables(template_str, value_map, transformer=None):
  """
  Expand a template string like "blah blah $FOO blah" using given value mapping.
  """
  if template_str is None:
    return None
  else:
    if transformer is None:
      transformer = lambda v: v
    try:
      transformed_value_map = {k: transformer(v) for (k, v) in value_map.iteritems()}
      return Template(template_str).substitute(transformed_value_map)
    except Exception as e:
      raise ValueError("could not expand variable names in command '%s': %s" % (template_str, e))


def shell_expand_variables(template_str, value_map):
  """
  Expand a shell template string like "cp $SOURCE $TARGET/blah", also quoting values as needed
  to ensure shell safety.
  """
  return expand_variables(template_str, value_map, transformer=pipes.quote)


def shell_expand_to_popen(template, values):
  """
  Expand a template like "cp $SOURCE $TARGET/blah" into a list of popen arguments.
  """
  return [expand_variables(item, values) for item in shlex.split(template)]


#
# ---- File operations ----

def move_to_backup(path, backup_suffix=BACKUP_SUFFIX):
  """
  Move the given file or directory to the same name, with a backup suffix.
  If backup_suffix not supplied, move it to the extension ".bak".
  NB: If backup_suffix is supplied and is None, don't do anything.
  """
  if backup_suffix and os.path.exists(path):
    backup_path = path + backup_suffix
    # Some messy corner cases need to be handled for existing backups.
    # TODO: Note if this is a directory, and we do this twice at once, there is a potential race
    # that could leave one backup inside the other.
    if os.path.islink(backup_path):
      os.unlink(backup_path)
    elif os.path.isdir(backup_path):
      shutil.rmtree(backup_path)
    shutil.move(path, backup_path)


def make_all_dirs(path, mode=0777):
  """
  Ensure local dir, with all its parent dirs, are created.
  Unlike os.makedirs(), will not fail if the path already exists.
  """
  if not os.path.isdir(path):
    os.makedirs(path, mode=mode)


def make_parent_dirs(path):
  """
  Ensure parent directories of a file are created as needed.
  """
  dir = os.path.dirname(path)
  if dir and not os.path.isdir(dir):
    os.makedirs(dir)
  return path


@contextmanager
def atomic_output_file(dest_path, make_parents=False, backup_suffix=None, suffix=".partial.%s"):
  """
  A context manager for convenience in writing a file or directory in an atomic way. Set up
  a temporary name, then rename it after the operation is done, optionally making a backup of
  the previous file or directory, if present.
  """
  tmp_path = ("%s" + suffix) % (dest_path, new_uid())
  if make_parents:
    make_parent_dirs(tmp_path)
  yield tmp_path
  if not os.path.exists(tmp_path):
    raise IOError("failure in writing file '%s': target file '%s' missing" % (dest_path, tmp_path))
  if backup_suffix:
    move_to_backup(dest_path, backup_suffix=backup_suffix)
  # If the target already exists, and is a directory, it has to be removed.
  if os.path.isdir(dest_path):
    shutil.rmtree(dest_path)
  shutil.move(tmp_path, dest_path)


def read_string_from_file(path):
  """
  Read entire contents of file into a string.
  """
  with open(path, "rb") as f:
    value = f.read()
  return value


def write_string_to_file(path, string, make_parents=False, backup_suffix=BACKUP_SUFFIX):
  """
  Write entire file with given string contents, atomically. Keeps backup by default.
  """
  with atomic_output_file(path, make_parents=make_parents, backup_suffix=backup_suffix) as tmp_path:
    with open(tmp_path, "wb") as f:
      f.write(string)


def set_file_mtime(path, mtime, atime=None):
  """Set access and modification times on a file."""
  if not atime:
    atime = mtime
  f = file(path, 'a')
  try:
    os.utime(path, (atime, mtime))
  finally:
    f.close()


def copyfile_atomic(source_path, dest_path, make_parents=False, backup_suffix=None):
  """
  Copy file on local filesystem in an atomic way, so partial copies never exist. Preserves timestamps.
  """
  with atomic_output_file(dest_path, make_parents=make_parents, backup_suffix=backup_suffix) as tmp_path:
    shutil.copyfile(source_path, tmp_path)
    set_file_mtime(tmp_path, os.path.getmtime(source_path))


def copytree_atomic(source_path, dest_path, make_parents=False, backup_suffix=None, symlinks=False):
  """
  Copy a file or directory recursively, and atomically, reanaming file or top-level dir when done.
  Unlike shutil.copytree, this will not fail on a file.
  """
  if os.path.isdir(source_path):
    with atomic_output_file(dest_path, make_parents=make_parents, backup_suffix=backup_suffix) as tmp_path:
      shutil.copytree(source_path, tmp_path, symlinks=symlinks)
  else:
    copyfile_atomic(source_path, dest_path, make_parents=make_parents, backup_suffix=backup_suffix)


def movefile(source_path, dest_path, make_parents=False, backup_suffix=None):
  """
  Move file. With a few extra options.
  """
  if make_parents:
    make_parent_dirs(dest_path)
  move_to_backup(dest_path, backup_suffix=backup_suffix)
  shutil.move(source_path, dest_path)


def rmtree_or_file(path, ignore_errors=False, onerror=None):
  """
  rmtree fails on files or symlinks. This removes the target, whatever it is.
  """
  if os.path.isdir(path) and not os.path.islink(path):
    shutil.rmtree(path, ignore_errors=ignore_errors, onerror=onerror)
  else:
    os.unlink(path)


def chmod_native(path, mode_expression, recursive=False):
  """
  This is ugly and will only work on POSIX, but the built-in Python os.chmod support
  is very minimal, and neither supports fast recursion nor "+X" type expressions, both
  of which are slow/costly for large trees. So just shell out.
  """
  popenargs = ["chmod"]
  if recursive:
    popenargs.append("-R")
  popenargs.append(mode_expression)
  popenargs.append(path)
  subprocess.check_call(popenargs)


def file_sha1(path):
  """
  Compute SHA1 hash of a file.
  """
  sha1 = hashlib.sha1()
  with open(path, "rb") as f:
    while True:
      block = f.read(2 ** 10)
      if not block:
        break
      sha1.update(block)
    return sha1.hexdigest()



"""
A simple utility to log calls to functions, for debugging.
"""

__author__ = 'jlevy'

import functools
import logging
from logging import log


def log_calls_with(severity):
  """Create a decorator to log calls and return values of any function, for debugging."""

  def decorator(fn):
    @functools.wraps(fn)
    def wrap(*params, **kwargs):
      call_str = "%s(%s)" % (
        fn.__name__, ", ".join([repr(p) for p in params] + ["%s=%s" % (k, repr(v)) for (k, v) in kwargs.items()]))
      # TODO: Extract line number from caller and use that in logging.
      log(severity, ">> %s", call_str)
      ret = fn(*params, **kwargs)
      # TODO: Add a way to make return short or omitted.
      log(severity, "<< %s: %s", call_str, repr(ret))
      return ret

    return wrap

  return decorator

# Convenience decorators for logging.
log_calls_info = log_calls_with(logging.INFO)
log_calls = log_calls_with(logging.DEBUG)



"""
Instaclone configuration handling.
"""

__author__ = 'jlevy'

import logging as log
import os
import re
import sys
from collections import namedtuple, OrderedDict

from enum import Enum  # enum34 pip
import yaml  # PyYAML pip
from functools32 import lru_cache  # functools32 pip

from log_calls import log_calls

import utils

_required_fields = "local_path remote_path remote_prefix copy_type upload_command download_command"
_other_fields = "version version_hashable version_command"

ConfigBase = namedtuple("ConfigBase", _other_fields + " " + _required_fields)

CONFIGS_REQUIRED = _required_fields.split()

CONFIG_VERSION_RE = re.compile("^[\\w.-]+$")


def _stringify_config_field(value):
  return value.name if isinstance(value, Enum) else str(value)


class Config(ConfigBase):
  """Configuration for a single item."""

  def as_string_dict(self):
    d = dict(self._asdict())
    return {k: _stringify_config_field(v) for (k, v) in d.iteritems() if v is not None}


CONFIG_NAME = "instaclone"
CONFIG_DIR_ENV = "INSTACLONE_DIR"
CONFIG_HOME_DIR = ".instaclone"

DEFAULT_ITEM_NAME = "default"


class ConfigError(RuntimeError):
  pass


CopyType = Enum("CopyType", "copy symlink hardlink")


@lru_cache(maxsize=None)
def _locate_config_dir():
  """Check for which config directory to use."""
  if CONFIG_DIR_ENV in os.environ:
    config_dir = os.environ[CONFIG_DIR_ENV]
  else:
    config_dir = os.path.join(os.environ["HOME"], CONFIG_HOME_DIR)
  return config_dir


def _locate_config_file(search_dirs):
  """Look in common locations for config file."""
  tried = []
  for base in search_dirs:
    for path in [os.path.join(base, CONFIG_NAME + suffix) for suffix in ".yml", ".json"]:
      log.debug("searching for config file: %s", path)
      tried.append(path)
      if os.path.isfile(path):
        log.info("using config file: %s", path)
        return path
  raise ConfigError("no config file found in: %s" % ", ".join(tried))


@log_calls
def _load_raw_configs(override_path):
  """Find first config in override_path, current dir, or config dir."""
  if override_path:
    path = override_path
  else:
    search_dirs = [".", _locate_config_dir()]
    path = _locate_config_file(search_dirs)

  with open(path) as f:
    parsed_configs = yaml.safe_load(f)

  out = []
  try:
    items = parsed_configs["items"]
    for config_dict in items:
      # TODO: Could overlay global configs and local ones here.
      combined = {key: None for key in Config._fields}
      combined.update(config_dict)

      try:
        out.append(combined)
      except TypeError as e:
        raise ConfigError("error in config value: %s: %s" % (e, config_dict))
  except ValueError as e:
    raise ConfigError("error reading config file: %s" % e)

  return out


def _parse_validate(raw_config_list):
  for raw in raw_config_list:
    # Validation.
    for key in CONFIGS_REQUIRED:
      if key not in raw or raw[key] is None:
        raise ConfigError("must specify '%s' in item config: %s" % (key, raw))

    if "version" in raw and not CONFIG_VERSION_RE.match(str(raw["version"])):
      raise ConfigError("invalid version string: '%s'" % raw["version"])
    if "version" not in raw and "version_hashable" not in raw and "version_command" not in raw:
      raise ConfigError("must specify 'version', 'version_hashable', or 'version_command' in item config: %s" % raw)

    # Validate shell templates.
    # For these, we don't expand environment variables here, but instead do it at once at call time.
    for key in "upload_command", "download_command":
      try:
        utils.shell_expand_to_popen(raw[key], {"REMOTE": "dummy", "LOCAL": "dummy"})
      except ValueError as e:
        raise ConfigError("invalid command in config value for %s: %s" % (key, e))

    # Normalize and expand environment variables.
    for key in "local_path", "remote_prefix", "remote_path":
      if key.startswith("/"):
        raise ConfigError("currently only support relative paths for local_path and remote_path: %s" % key)
      raw[key] = raw[key].rstrip("/")

      try:
        raw[key] = utils.expand_variables(raw[key], os.environ)
      except ValueError as e:
        raise ConfigError("invalid command in config value for %s: %s" % (key, e))

    # Parse enums.
    try:
      raw["copy_type"] = CopyType[raw["copy_type"]]
    except KeyError:
      raise ConfigError("invalid copy type: %s" % raw["copy_type"])

    yield Config(**raw)


@log_calls
def set_up_cache_dir():
  config_dir = _locate_config_dir()
  cache_dir = os.path.join(config_dir, "cache")
  if not os.path.exists(cache_dir):
    log.info("cache dir not found, so creating: %s", cache_dir)
    utils.make_all_dirs(cache_dir)
  return cache_dir


def load(override_path=None):
  """Load all configs from a single file. Use override_path or the first one found in standard locations."""
  return list(_parse_validate(_load_raw_configs(override_path)))


def print_configs(configs, stream=sys.stdout):
  yaml.dump({"items": [config.as_string_dict() for config in configs]},
            stream=stream, default_flow_style=False)


def _yaml_ordering_support():
  """
  Get yaml lib to handle OrderedDicts.
  See http://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
  """
  _mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG

  def dict_representer(dumper, data):
    return dumper.represent_dict(data.iteritems())

  def dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))

  yaml.add_representer(OrderedDict, dict_representer)
  yaml.add_constructor(_mapping_tag, dict_constructor)


_yaml_ordering_support()



#!/usr/bin/env python
"""
For further documentation, see: https://github.com/jlevy/instaclone
"""

from __future__ import print_function

import logging as log
import argparse
import sys

NAME = "instaclone"
VERSION = "0.1.11"
DESCRIPTION = "instaclone: Fast, cached installations of versioned files"
LONG_DESCRIPTION = __doc__

LOG_STREAM = sys.stderr


def log_setup(level):
  if level == log.DEBUG:
    log.basicConfig(format="%(levelname).1s %(filename)16s:%(lineno)-4d  %(message)s", level=level,
                    stream=LOG_STREAM)
  else:
    log.basicConfig(format="%(message)s", level=level, stream=LOG_STREAM)

    def brief_excepthook(exctype, value, traceback):
      print("error: %s" % value, file=sys.stderr)
      print("(run with --debug for traceback info)", file=sys.stderr)
      sys.exit(2)

    sys.excepthook = brief_excepthook


def main():
  import instaclone
  import configs

  parser = argparse.ArgumentParser(description=DESCRIPTION, version=VERSION, epilog="\n" + __doc__,
                                   formatter_class=argparse.RawTextHelpFormatter)
  parser.add_argument("command", help="%s command" % NAME, choices=instaclone._command_list)
  parser.add_argument("--config", help="config YAML or JSON file to override usual search path")
  parser.add_argument("-f", "--force",
                      help="force operation, clobbering any existing cached or local targets (use with care)",
                      action="store_true")
  parser.add_argument("--debug", help="enable debugging output", action="store_true")
  args = parser.parse_args()

  log_setup(log.DEBUG if args.debug else log.INFO)

  config_list = configs.load(override_path=args.config)

  instaclone.run_command(instaclone.Command[args.command], config_list, force=args.force)


if __name__ == '__main__':
  main()



"""
Instaclone main library.
"""

from __future__ import print_function

__author__ = 'jlevy'

import logging as log
import re
import sys
import os
import shutil

from enum import Enum  # enum34

# The subprocess module has known threading issues, so prefer subprocess32.
try:
  import subprocess32 as subprocess
except ImportError:
  import subprocess

import configs
from utils import atomic_output_file, write_string_to_file, DEV_NULL
from utils import move_to_backup, movefile, copytree_atomic, rmtree_or_file, file_sha1
from utils import make_all_dirs, make_parent_dirs, chmod_native
from utils import shell_expand_to_popen
from utils import dict_merge
from log_calls import log_calls

SHELL_OUTPUT = sys.stderr


class AppError(RuntimeError):
  pass


@log_calls
def _make_readonly(path, silent=False):
  if silent and not os.path.exists(path):
    return
  return chmod_native(path, "ugo-w", recursive=True)


@log_calls
def _make_writable(path, silent=False):
  if silent and not os.path.exists(path):
    return
  return chmod_native(path, "u+w", recursive=True)


def _upload_file(command_template, local_path, remote_loc):
  popenargs = shell_expand_to_popen(command_template,
                                    dict_merge(os.environ, {"REMOTE": remote_loc, "LOCAL": local_path}))
  log.info("uploading: %s", " ".join(popenargs))
  # TODO: Find a way to support force here (e.g. add or remove -f to s4cmd)
  subprocess.check_call(popenargs, stdout=SHELL_OUTPUT, stderr=SHELL_OUTPUT, stdin=DEV_NULL)


def _download_file(command_template, remote_loc, local_path):
  with atomic_output_file(local_path, make_parents=True) as temp_target:
    popenargs = shell_expand_to_popen(command_template,
                                      dict_merge(os.environ, {"REMOTE": remote_loc, "LOCAL": temp_target}))
    log.info("downloading: %s", " ".join(popenargs))
    # TODO: Find a way to support force here.
    subprocess.check_call(popenargs, stdout=SHELL_OUTPUT, stderr=SHELL_OUTPUT, stdin=DEV_NULL)

# For simplicity, we currently only support zip compression.
# We use command-line standard zip/unzip instead of Python zip, since it is a bit more performant
# and makes the archiving mechanism pluggable in the future.
# TODO: Note zip/unzip by default follows symlinks, so full contents are included. Consider making this a flag.

ARCHIVE_SUFFIX = ".zip"


def _compress_dir(local_dir, archive_path, force=False):
  if os.path.exists(archive_path):
    if force:
      log.info("deleting previous archive: %s", archive_path)
      os.unlink(archive_path)
    else:
      raise AppError("archive already in cache (has version changed?): %s" % archive_path)
  with atomic_output_file(archive_path) as temp_archive:
    make_parent_dirs(temp_archive)
    popenargs = shell_expand_to_popen("zip -q -r $ARCHIVE $DIR", {"ARCHIVE": temp_archive, "DIR": "."})
    cd_to = local_dir
    log.debug("using cwd: %s", cd_to)
    log.info("compress: %s", " ".join(popenargs))
    subprocess.check_call(popenargs, cwd=cd_to, stdout=SHELL_OUTPUT, stderr=SHELL_OUTPUT, stdin=DEV_NULL)


def _decompress_dir(archive_path, target_path, force=False):
  if os.path.exists(target_path):
    if force:
      log.info("deleting previous dir: %s", target_path)
      shutil.rmtree(target_path)
    else:
      raise AppError("target already exists: %s" % target_path)
  with atomic_output_file(target_path) as temp_dir:
    popenargs = shell_expand_to_popen("unzip -q $ARCHIVE", {"ARCHIVE": archive_path, "DIR": temp_dir})
    make_all_dirs(temp_dir)
    cd_to = temp_dir
    log.debug("using cwd: %s", cd_to)
    log.info("decompress: %s", " ".join(popenargs))

    subprocess.check_call(popenargs, cwd=cd_to, stdout=SHELL_OUTPUT, stderr=SHELL_OUTPUT, stdin=DEV_NULL)


@log_calls
def _install_from_cache(cache_path, target_path, copy_type, force=False, make_backup=False):
  """
  Install a file or directory from cache, either symlinking, hardlinking, or copying.
  """

  def checked_remove():
    # Never backup links (they are probably previous installs).
    if os.path.islink(target_path):
      os.unlink(target_path)
    elif os.path.exists(target_path):
      if force:
        if make_backup:
          move_to_backup(target_path)
        else:
          rmtree_or_file(target_path)
      else:
        raise AppError("target already exists: %s" % target_path)

  if not os.path.exists(cache_path):
    raise AssertionError("cached file missing: %s" % cache_path)
  if copy_type == configs.CopyType.symlink:
    checked_remove()
    os.symlink(cache_path, target_path)
  elif copy_type == configs.CopyType.hardlink:
    if os.path.isdir(cache_path):
      raise AppError("can't hardlink a directory: %s" % cache_path)
    checked_remove()
    os.link(cache_path, target_path)
  elif copy_type == configs.CopyType.copy:
    checked_remove()
    copytree_atomic(cache_path, target_path)
  else:
    raise AssertionError("invalid copy_type: %s" % copy_type)


VERSION_SEP = ".$"
VERSION_END = "$"


class FileCache(object):
  """
  Manage uploading and downloading files to/from the cloud using a local cache to maintain copies.
  Also seamlessly support directories by archiving them as compressed files.
  The cache is not bounded and must be managed/cleaned up manually.
  """

  version = "1"

  def __init__(self, root_path):
    self.root_path = root_path.rstrip("/")
    self.contents_path = os.path.join(root_path, "contents")
    self.version_path = os.path.join(root_path, "version")
    self.setup_done = False
    assert os.path.exists(self.root_path)

  def setup(self):
    """Lazy initialize file cache post instantiation."""
    if not self.setup_done:
      if os.path.exists(self.version_path):
        log.info("using cache: %s", self.root_path)
      else:
        log.info("initializing new cache: %s", self.root_path)
        make_all_dirs(self.contents_path)
        write_string_to_file(self.version_path, FileCache.version + "\n")
      self.setup_done = True

  def __str__(self):
    return "FileCache@%s" % self.root_path

  def __repr__(self):
    return self.__str__()

  @staticmethod
  def versioned_path(config, version, suffix=""):
    return os.path.join(config.remote_path,
                        "%s%s%s%s" % (config.local_path, VERSION_SEP, version, VERSION_END),
                        "%s%s" % (os.path.basename(config.local_path), suffix))

  @staticmethod
  def pathify_remote_loc(remote_loc):
    return os.path.join(*re.findall("[a-zA-Z0-9_.-]+", remote_loc))

  def cache_path(self, config, version, suffix=""):
    return os.path.join(self.contents_path,
                        self.pathify_remote_loc(config.remote_prefix),
                        self.versioned_path(config, version, suffix))

  def remote_loc(self, config, version, suffix=""):
    return os.path.join(config.remote_prefix,
                        self.versioned_path(config, version, suffix))

  def _upload(self, config, cached_path, version):
    _upload_file(config.upload_command, cached_path, self.remote_loc(config, version))

  @log_calls
  def publish(self, config, version, force=False):
    # As precaution for users, we keep unarchived items in cache that may be symlinked to as read-only.
    cached_path = self.cache_path(config, version)
    try:
      _make_writable(cached_path, silent=True)
      self._publish_writable(config, version, force=force)
    finally:
      _make_readonly(cached_path, silent=True)

  def _publish_writable(self, config, version, force=False):
    local_path = config.local_path
    cached_path = self.cache_path(config, version)

    if os.path.islink(local_path):
      raise AppError("cannot publish symlinks (is path already published?): %s" % local_path)

    self.setup()

    # Directories are archived. Files are published as is.
    if os.path.isdir(local_path):
      cached_archive = self.cache_path(config, version, ARCHIVE_SUFFIX)
      remote_loc = self.remote_loc(config, version, ARCHIVE_SUFFIX)

      # We archive and then unarchive, to make sure we expand symlinks exactly the way
      # a future installation would (using zip/unzip).
      # TODO: This is usually what we want (think of relative symlinks like ../../foo), but we could make it an option.
      log.info("installing to cache: %s -> %s", local_path, cached_path)
      _compress_dir(local_path, cached_archive, force=force)
      _upload_file(config.upload_command, cached_archive, remote_loc)
      _decompress_dir(cached_archive, cached_path, force=force)
      # Leave the previous version of the tree as a backup.
      log.debug("installing back from cache: %s <- %s", local_path, cached_path)
      _install_from_cache(cached_path, local_path, config.copy_type, force=True, make_backup=True)
      log.info("published directory archive: %s -> %s", config.local_path, remote_loc)
    elif os.path.isfile(local_path):
      remote_loc = self.remote_loc(config, version)

      log.info("installing to cache: %s -> %s", local_path, cached_path)
      # For speed on large files, move it rather than copy.
      # Also make it read-only, just as it will be after install.
      movefile(local_path, cached_path, make_parents=True)
      _upload_file(config.upload_command, cached_path, remote_loc)
      log.debug("installing back from cache: %s <- %s", local_path, cached_path)
      _install_from_cache(cached_path, local_path, config.copy_type, force=False, make_backup=False)
      log.info("published file: %s -> %s", config.local_path, remote_loc)
    elif os.path.exists(local_path):
      # TODO: Consider handling symlinks.
      raise ValueError("only files or directories supported: %s" % local_path)
    else:
      raise ValueError("file not found: %s" % local_path)

  @log_calls
  def install(self, config, version, force=False):
    self.setup()
    cached_path = self.cache_path(config, version)
    if os.path.exists(cached_path):
      # It's a cached file or a cached directory and we've already unpacked it.
      log.info("installing from cache: %s <- %s", config.local_path, cached_path)
      _install_from_cache(cached_path, config.local_path, config.copy_type, force=force)
    else:
      # First try it as a directory/archive.
      remote_archive_loc = self.remote_loc(config, version, ARCHIVE_SUFFIX)
      cached_archive_path = self.cache_path(config, version, suffix=ARCHIVE_SUFFIX)
      is_dir = True
      # TODO: This could be cleaner, but it's nice to be data-driven and not require a config saying it's a dir or file.
      log.debug("checking if it's a directory by seeing if archive suffix exits")
      try:
        _download_file(config.download_command, remote_archive_loc, cached_archive_path)
      except subprocess.CalledProcessError:
        log.debug("doesn't look like an archived directory, so treating it as a file")
        is_dir = False
      if is_dir:
        log.info("installing directory: %s <- %s <- %s", config.local_path, cached_path, remote_archive_loc)
        _decompress_dir(cached_archive_path, cached_path, force=force)
      else:
        remote_loc = self.remote_loc(config, version)
        log.info("installing file: %s <- %s <- %s", config.local_path, cached_path, remote_loc)
        _download_file(config.download_command, remote_loc, cached_path)

      _make_readonly(cached_path)
      _install_from_cache(cached_path, config.local_path, config.copy_type, force=force)

  @log_calls
  def purge(self):
    log.info("purging cache: %s", self.root_path)
    _make_writable(self.root_path, silent=True)
    shutil.rmtree(self.root_path, ignore_errors=True)


def version_for(config):
  """
  The version for an item is either the explicit version specified by the user, or the SHA1 hash of hashable file.
  """
  bits = []
  if config.version:
    bits.append(str(config.version))
  if config.version_hashable:
    log.debug("computing sha1 of: %s", config.version_hashable)
    bits.append(file_sha1(config.version_hashable))
  if config.version_command:
    log.debug("version command: %s", config.version_command)
    popenargs = shell_expand_to_popen(config.version_command, os.environ)
    output = subprocess.check_output(popenargs, stderr=SHELL_OUTPUT, stdin=DEV_NULL).strip()
    if not configs.CONFIG_VERSION_RE.match(output):
      raise configs.ConfigError("invalid version output from version command: '%s'" % output)
    bits.append(output)

  return "-".join(bits)

#
# ---- Command line ----

Command = Enum("Command", "publish install purge configs")
_command_list = [c.name for c in Command]


def run_command(command, config_list, force=False):
  # Don't initialize cache for nondestructive commands.
  if command == Command.configs:
    configs.print_configs(config_list)
  else:
    cache_dir = configs.set_up_cache_dir()
    file_cache = FileCache(cache_dir)

    if command == Command.publish:
      for config in config_list:
        file_cache.publish(config, version_for(config), force=force)

    elif command == Command.install:
      for config in config_list:
        file_cache.install(config, version_for(config), force=force)

    elif command == Command.purge:
      file_cache.purge()

    else:
      raise AssertionError("unknown command: " + command)

# TODO:
# - failover_command that is executed if install fails, and failover_publish
# - failover_publish flag
# - merge command-line args and config file values?
# - --no-cache option that just downloads
# - let "instaclone purge" work even if not in a dir with a config file
# - think how to auto-purge all but one resource per branch (say) (and generalize to env variable)
# - "clean" command that deletes local resources (requiring -f if not in cache)
# - "unpublish" command that deletes a remote resource (and purges from cache)
# - command to unpublish all but most recent n versions of a resource
# - support compressing files as well as archives
# - consider a pax-based hardlink tree copy option (since pax is cross platform, unlike cp's options)
# - init command to generate a config
# - "--offline" mode for install (i.e. will fail if it has to download)
# - test out more custom transport commands (s3cmd, awscli, wget, etc.)
# - for the custom transport like curl, figure out handling of shell redirects (or just require)
